#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

const int k=4;
const char alphabet[4]    = {'A','C','G','T'};
const int  vals[4]        = { 0,  1,  2,  3 };
const int  complements[4] = { 3,  2,  1,  0 };

/* based on code by Roy van Rijn, 
 * http://www.redcode.nl/blog/2011/02/de-bruijn-sequence-in-constant-amortized-time
 * and similar code on the wikipedia page for deBruijn sequences.
 */
void generateLyndonWords(int t, int p, int k, int *a, int alen, int *seq, int *seqlen) {
   if (t > alen-1) {
       if((alen-1)%p==0) {
           for(int i = 1; i < p+1; i++) {
              seq[(*seqlen)] = a[i];
              (*seqlen)++;
           }
       }
   } else {
       a[t] = a[t-p];
       generateLyndonWords(t+1,p, k, a, alen, seq, seqlen);
       for(int j=a[t-p]+1; j<=k-1; j++) {
           a[t] = j;
           generateLyndonWords(t+1,t, k, a, alen, seq, seqlen);
       }
   }
} 

int sequenceToNum(int *seq, int k, int n) {
    assert(n < 16);
        
    int num=0; 
    for (int i=0; i<n; i++) {
        num *= k;
        num += seq[i];
    }

    return num;
}

void numToSequence(const int num, const int k, const int n, int *seq) {
    int mynum = num;
    int select = (int)pow((double)k, (double)(n-1));
    for (int digit=0; digit<n; digit++) {
        seq[digit] = (mynum / select);
        mynum = mynum % select;
        select /= k;
    }

    return;
}

void sequenceToString(int *seq, int n, char *str) {
    for (int digit=0; digit<n; digit++)  {
//        assert(seq[digit] < k);
        assert(seq[digit] > -1);
        str[digit] = alphabet[seq[digit]];  
    }
    return;
}

void numToString(const int num, const int k, const int n, char *str) {
    int digits[n];
    numToSequence(num, k, n, digits);
    sequenceToString(digits, n, str);

    return;
}

int reverseComplement(int kmer, int k, int n) {
    int rc = 0;
    int select = 1;
    for (int digit=0; digit < n; digit++) {
        rc *= k;
        rc += complements[ (kmer / select) % k ];
        kmer = (kmer / k) * k;
        select *= k;
    }

    return rc;
}

void generateDeBruijnSequence(int k, int n, int **nmerseq) {
     int maxlen = (int)pow((double)k,(double)n);
     int alen=n+1;
     int *sequence = (int *)malloc((maxlen+n-1)*sizeof(int));
     int *a = (int *)malloc(alen*sizeof(int));
     int slen=0;

     for (int i=0; i<alen; i++) a[i] = 0;
     generateLyndonWords(1, 1, k, a, alen, sequence, &slen);
     free(a);

     *nmerseq = (int *)malloc( (slen)*sizeof(int) );

     for (int i=0; i<n-1; i++) {
        sequence[slen] = sequence[i];
        slen++;
     }

     for (int i=0; i<slen-n+1; i++) {
        (*nmerseq)[i] =  sequenceToNum(&(sequence[i]),k,n);
     }  

     free(sequence);
}

void partitionDeBruijnSequence(const int k, const int m, const int nproc, const int *seq, 
                               int **lookup) {
    int nseq=(int)pow((double)k,(double)m);
    *lookup = (int *)malloc(nseq * sizeof(int)); 
    const int UNASSIGNED=-1;
        
    for (int i=0; i<nseq; i++)
        (*lookup)[i] = UNASSIGNED;

    int nindseq = 0;  /* independant sequences */
    for (int i=0; i<nseq; i++) {
        int s   = seq[i];
        int rcs = reverseComplement(s,k,m); 
        int lowest = s;
        if (rcs < s) lowest = rcs;
        if ((*lookup)[lowest] == UNASSIGNED) {
            (*lookup)[s]   = nindseq;
            (*lookup)[rcs] = nindseq;
            nindseq++;
        }
    }

    int start[nproc];
    int end[nproc];
    start[0] = 0;
    for (int p=0; p<nproc; p++) {
        end[p] = start[p] + (nindseq+p)/nproc;
        start[p+1] = end[p];
    }

    for (int i=0; i<nseq; i++) {
        int s = (*lookup)[seq[i]];
        int p=0;
        while ((p<nproc) && !((s >= start[p]) && (s < end[p])) )
            p++;
        assert(p < nproc);
        (*lookup)[seq[i]] = p;
    }

    int counts[nproc];
    for (int i=0; i<nproc; i++) 
        counts[i]=0;

    for (int i=0; i<nseq; i++) {
        counts[(*lookup)[seq[i]]]++;
    }

    printf("Load balance: ");
    for (int i=0; i<nproc; i++)
        printf("%2d:%d ", i, counts[i]);
    printf("\n");

    return;
}

void neighbours(int seq, int k, int m, int nNeighBases, int maxnneighs, int *nneighs, int *neighs) {
    int nneighseq=(int)pow((double)k, (double)nNeighBases);
    *nneighs = (nNeighBases+1)*nneighseq;
    assert(*nneighs <= maxnneighs);

    for (int i=0; i<=nNeighBases; i++) {     /* # of bases before seq */
        int filter = (int)pow((double)k,(double)(nNeighBases-i));
        int rightshift = (int)pow((double)k, (double)(m+nNeighBases-i));
        for (int j=0; j<nneighseq; j++) {   /* sequence to prepend+postpen */
            int prefix = j / filter;
            int suffix = j % filter;
            if (filter == 1) suffix = 0;

            neighs[i*nneighseq+j] = prefix*rightshift + seq*filter + suffix;
        }
    }
}

void partitionDeBruijnSequenceByNeighbours(const int k, const int m, const int nproc, 
                            const int nNeighBases, const int *seq, int **lookup) {

    int nshortseq=(int)pow((double)k,(double)(m-nNeighBases));
    int nseq=(int)pow((double)k,(double)(m));
    int nneighs=(nNeighBases+1)*(int)pow((double)k,(double)(nNeighBases));
    *lookup = (int *)malloc(nseq * sizeof(int)); 
    int *neighs = (int *)malloc(nneighs * sizeof(int)); 
    const int UNASSIGNED=-1;
    int outnneighs;
        
    for (int i=0; i<nseq; i++)
        (*lookup)[i] = UNASSIGNED;

    int nindseq = 0;  /* independent sequences */
    for (int i=0; i<nshortseq; i++) {
        int ss  = seq[i];
        neighbours(ss, k, m-nNeighBases, nNeighBases, nneighs, &outnneighs, neighs);
        for (int j=0; j<outnneighs; j++) {
            int s = neighs[j];
            int rcs = reverseComplement(s,k,m); 
            int lowest = s;
            if (rcs < s) lowest = rcs;

            if ((*lookup)[lowest] == UNASSIGNED) {
                (*lookup)[s]   = nindseq;
                (*lookup)[rcs] = nindseq;
                nindseq++;
            }

        }
    }
    free(neighs);

    for (int i=0; i<nseq; i++)
        assert((*lookup)[i] != UNASSIGNED);

    int start[nproc];
    int end[nproc];
    start[0] = 0;
    for (int p=0; p<nproc; p++) {
        end[p] = start[p] + (nindseq+p)/nproc;
        start[p+1] = end[p];
    }

    for (int i=0; i<nseq; i++) {
        int s = (*lookup)[i];
        int p=0;
        while ((p<nproc) && !((s >= start[p]) && (s < end[p])) )
            p++;
        assert(p < nproc);
        (*lookup)[i] = p;
    }

    int counts[nproc];
    for (int i=0; i<nproc; i++) 
        counts[i]=0;

    for (int i=0; i<nseq; i++) {
        assert((*lookup)[i] >= 0 && (*lookup)[i] < nproc);
        counts[(*lookup)[i]]++;
    }

    printf("Load balance: ");
    for (int i=0; i<nproc; i++)
        printf("%2d:%d ", i, counts[i]);
    printf("\n");

    return;
}

int *generateLookupTable(const int k, const int m, const int nproc) {
    int *mmerseq;
    int *lookup;
    generateDeBruijnSequence(k,m,&mmerseq);    
    partitionDeBruijnSequence(k,m,nproc,mmerseq,&lookup);
    free(mmerseq);
    return lookup;
}

int *generateLookupTableByNeighbours(const int k, const int m, const int nNeighBases, const int nproc) {
    int *mmerseq;
    int *lookup;
    generateDeBruijnSequence(k,m-nNeighBases,&mmerseq);    
    partitionDeBruijnSequenceByNeighbours(k,m,nproc,nNeighBases,mmerseq,&lookup);
    free(mmerseq);
    return lookup;
}
